﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace ParallelDemo
{
    class Program
    {
        static void Main(string[] args)
        {
            // Create a large list of numbers
            var numbers = Enumerable.Range(0, 5000000).ToList();

            Utilities.Timings("Sequential", () =>
            {
                Console.WriteLine("Number of primes = {0}", GetNumberOfPrimes(numbers));
            });
            
            // Get the total number of primes, in a parallel fashion using .NET 3.5
            Utilities.Timings("Parallel 2.0/3.5", () =>
            {
                Console.WriteLine("Number of primes = {0}", GetNumberOfPrimes_NET20(numbers));
            });

            Utilities.Timings("Parallel 4", () =>
            {
                Console.WriteLine("Number of primes = {0}", GetNumberOfPrimes_NET4(numbers));
            });

            Utilities.Timings("Parallel 4 Best", () =>
            {
                Console.WriteLine("Number of primes = {0}", GetNumberOfPrimes_NET4_Best(numbers));
            });

            #region Declarative Versions

            Utilities.Timings("Sequential LINQ - Declarative Programming", () =>
            {
                int numberOfPrimes = numbers.Where(n => Utilities.IsPrime(n)).Count();
                Console.WriteLine("Number of primes = {0}", numberOfPrimes);
            });

            Utilities.Timings("Parallel LINQ - Declarative Data Parallelism", () =>
            {
                // Implement using both LINQ forms

               // int numberOfPrimes = numbers.AsParallel().Where(n => Utilities.IsPrime(n)).Count();

                var temp = from n in numbers.AsParallel()
                           where Utilities.IsPrime(n)
                           select n;
                int numberOfPrimes = temp.Count();
                
                Console.WriteLine("Number of primes = {0}", numberOfPrimes);
            });

            #endregion

            Console.WriteLine();
            Console.WriteLine("Press any key to exit...");
            Console.ReadKey();
        }

        public static int GetNumberOfPrimes(IList<int> numbers)
        {
            int total = 0;
            foreach (var number in numbers)
            {
                if (Utilities.IsPrime(number))
                    ++total;
            }
            return total;
        }

        public static int GetNumberOfPrimes_NET20(IList<int> numbers)
        {
            long total = 0;

            MyParallelFor_NET20(0, numbers.Count, (i) =>
                {
                    if (Utilities.IsPrime(numbers[i]))
                        Interlocked.Increment(ref total);
                }
            );
            
            return Convert.ToInt32(total);
        }

        // Taken from Stephen Toub's "Patterns of Parallel Programming"
        // It partitions into even number of elements per core
        public static void MyParallelFor_NET20(int inclusiveLowerBound, int exclusiveUpperBound, Action<int> body)
        {
            int size = exclusiveUpperBound - inclusiveLowerBound;
            int numProcs = Environment.ProcessorCount;
            int range = size / numProcs;
            // Keep track of the number of threads remaining to complete.
            int remaining = numProcs;
            
            // Create a way to track when we're completed
            using (ManualResetEvent mre = new ManualResetEvent(false))
            {
                // Create each of the threads.
                for (int p = 0; p < numProcs; p++)
                {
                    int start = p * range + inclusiveLowerBound;
                    int end = (p == numProcs - 1) ? exclusiveUpperBound : start + range;
                    // Queue work for "block" of threads
                    ThreadPool.QueueUserWorkItem(delegate
                    {
                        // Execute our body
                        for (int i = start; i < end; i++) 
                            body(i);
                        if (Interlocked.Decrement(ref remaining) == 0) 
                            mre.Set();
                    });
                }
                
                // Wait for all threads to complete.
                mre.WaitOne();
            }
        }

        public static int GetNumberOfPrimes_NET4(IList<int> numbers)
        {
            int total = 0;

            Parallel.For(0, numbers.Count, number =>
            {
                if (Utilities.IsPrime(number))
                {
                    Interlocked.Increment(ref total);
                }
            });
            return total;
        }

        // In this case, this is only slightly better than above, since Interlocked doesn't hold a lock.
        // If locking were required, this would be dramatically better than previous version
        public static int GetNumberOfPrimes_NET4_Best(IList<int> numbers)
        {
            int total = 0;
            // Implement using Parallel.For<TLocal>
            Parallel.For(0, numbers.Count,
                // Start by initializing TLocal (int) to 0
                () => 0,
                // For each element, do work, and return the new local value
                (number, state, local) =>
                {
                    if (Utilities.IsPrime(number))
                        local++;
                    return local;
                },
                // Final delegate does merging, once per thread used
                (local) =>
                {
                    Interlocked.Add(ref total, local);
                });

            return total;
        }

    }
}
