﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ParallelDemo
{
    public partial class Form1 : Form
    {
        private List<int> numbers = Enumerable.Range(1, 10000000).ToList();

        public Form1()
        {
            InitializeComponent();
            this.timer1.Tick += (s, e) => this.labelCurrentTime.Text = "Current time: " + DateTime.Now.ToString("hh:mm:ss.ffff");
        }

        private void button1_Click(object sender, EventArgs e)
        {
            this.timer1.Enabled = true;
            this.button1.Enabled = false;
        }

        private void buttonDoWorkSynchronously_Click(object sender, EventArgs e)
        {
            this.labelTotalPrimes.Text = string.Format("Total Primes: {0}", GetNumberOfPrimes(numbers));
        }


        private int GetNumberOfPrimes(IList<int> numbers)
        {
            int total = 0;
            foreach (var number in numbers)
            {
                if (Utilities.IsPrime(number))
                    ++total;
            }
            return total;
        }

        private void buttonClear_Click(object sender, EventArgs e)
        {
            this.labelTotalPrimes.Text = "Total Primes: ";
        }
        private void DisableUI()
        {
            this.buttonDoWorkTasks.Enabled = false;
            this.buttonDoWorkTasks2.Enabled = false;
            this.buttonDoWorkBW.Enabled = false;
            this.buttonDoWorkThreadPool.Enabled = false;
            this.buttonDoWorkSynchronously.Enabled = false;
            this.buttonDoWorkThread.Enabled = false;
        }
        private void EnableUI()
        {
            this.buttonDoWorkTasks.Enabled = true;
            this.buttonDoWorkTasks2.Enabled = true;
            this.buttonDoWorkBW.Enabled = true;
            this.buttonDoWorkThreadPool.Enabled = true;
            this.buttonDoWorkSynchronously.Enabled = true;
            this.buttonDoWorkThread.Enabled = true;
        }

        #region Do Work Using Thread
        
        private void buttonDoWorkThread_Click(object sender, EventArgs e)
        {
            Thread newThread = new Thread(ThreadWork);
            DisableUI();
            newThread.Start();
        }
        private void ThreadWork()
        {
            int primes = GetNumberOfPrimes(numbers);
            this.BeginInvoke(new Action( () =>
                {
                    this.EnableUI();
                    this.labelTotalPrimes.Text = string.Format("Total Primes: {0}", primes);
                }));
        }

        #endregion

        #region Do Work Using ThreadPool

        private void buttonDoWorkThreadPool_Click(object sender, EventArgs e)
        {
            this.DisableUI();
            var context = SynchronizationContext.Current;
            ThreadPool.QueueUserWorkItem(_ =>
                {
                    int primes = GetNumberOfPrimes(numbers);
                    context.Post((s) =>
                        {
                            this.EnableUI();
                            this.labelTotalPrimes.Text = string.Format("Total Primes: {0}", primes);
                        }, null);
                });
        }

        #endregion

        #region BackgroundWorker

        private void buttonDoWorkBW_Click(object sender, EventArgs e)
        {
            this.progressBar1.Value = 0;
            this.progressBar1.Minimum = 0;
            this.progressBar1.Maximum = 100;
            this.DisableUI();
            this.backgroundWorker1.RunWorkerAsync(numbers);

        }

        private void backgroundWorker1_DoWork(object sender, DoWorkEventArgs e)
        {
            IList<int> numbers = e.Argument as IList<int>;
            int total = 0;
            for (int i = 0; i < numbers.Count; ++i)
            {
                // Report our progress every 500 items
                if (i % 500 == 0)
                    backgroundWorker1.ReportProgress((100 * i) / numbers.Count);

                if (Utilities.IsPrime(numbers[i]))
                    ++total;
            }
            
            // Store answer in e.Result
            e.Result = total;
        }

        private void backgroundWorker1_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            this.progressBar1.Value = e.ProgressPercentage;
        }

        private void backgroundWorker1_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            this.progressBar1.Value = 100;
            this.labelTotalPrimes.Text = string.Format("Total Primes: {0}", e.Result);
            this.EnableUI();
        }

        #endregion
        
        #region Tasks

        private void buttonDoWorkTasks_Click(object sender, EventArgs e)
        {
            int total = 0;

            this.DisableUI();
            Task.Factory.StartNew(() =>
            {
                total = this.GetNumberOfPrimes(numbers);
            }
            ).ContinueWith(_ =>
            {
                this.labelTotalPrimes.Text = string.Format("Total Primes: {0}", total);
                this.EnableUI();
            }, TaskScheduler.FromCurrentSynchronizationContext());
        }

        #endregion

        #region Tasks (Improved)

        private void buttonDoWorkTasks2_Click(object sender, EventArgs e)
        {
            var uiSched = TaskScheduler.FromCurrentSynchronizationContext();
            var ui = new TaskFactory(uiSched);
            int total = 0;

            this.DisableUI();

            Task.Factory.StartNew(() =>
            {
                for (int i = 0; i < numbers.Count; ++i)
                {
                    // Report our progress every 500 items
                    if (i % 500 == 0)
                    {
                        ui.StartNew(() => this.progressBar1.Value = (100 * i) / numbers.Count);
                    }

                    if (Utilities.IsPrime(numbers[i]))
                        ++total;
                }
            }).ContinueWith(_ =>
                {
                    this.labelTotalPrimes.Text = string.Format("Total Primes: {0}", total);
                    this.EnableUI();
                }, uiSched);
        }

        #endregion

    }
}
